#include "casertestreader.h"
#include "cucontrolsreader_abs.h"
#include <cumacros.h>
#include <cumbiapool.h>
#include <cudata.h>
#include <QContextMenuEvent>
#include <QMetaProperty>

#include "cucontrolsfactories_i.h"
#include "cucontrolsfactorypool.h"
#include "culinkstats.h"
#include "cucontextmenu.h"
#include "cucontext.h"

/** @private */
class CaSerTestReaderPrivate {
public:
    bool auto_configure;
    bool read_ok;
    CuContext *context;
    long  id;
    double creat_usecs;
};

CaSerTestReader::CaSerTestReader(QObject *parent,
                                 CumbiaPool *cumbia_pool,
                                 const CuControlsFactoryPool &fpool,
                                 double creat_usecs,
                                 long id) :
    QObject(parent), CuDataListener() {
    m_init();
    d->context = new CuContext(cumbia_pool, fpool);
    d->creat_usecs = creat_usecs;
    d->id = id;
}

void CaSerTestReader::m_init() {
    d = new CaSerTestReaderPrivate;
    d->context = NULL;
    d->auto_configure = true;
    d->read_ok = false;
}

CaSerTestReader::~CaSerTestReader() {
    pdelete("~CaSerTestReader %p", this);
    delete d->context;
    delete d;
}

QString CaSerTestReader::source() const {
    if(CuControlsReaderA* r = d->context->getReader())
        return r->source();
    return "";
}

/** \brief returns the pointer to the CuContext
 *
 * CuContext sets up the connection and is used as a mediator to send and get data
 * to and from the reader.
 *
 * @see CuContext
 */
CuContext *CaSerTestReader::getContext() const {
    return d->context;
}

long CaSerTestReader::creat_usecs() const {
    return d->creat_usecs;
}

long CaSerTestReader::id() const {
    return d->id;
}

/** \brief Connect the reader to the specified source.
 *
 * If a reader with a different source is configured, it is deleted.
 * If options have been set with QuContext::setOptions, they are used to set up the reader as desired.
 *
 * @see QuContext::setOptions
 * @see source
 */
void CaSerTestReader::setSource(const QString &s)
{
    CuControlsReaderA * r = d->context->replace_reader(s.toStdString(), this);
    if(r)
        r->setSource(s);
}

void CaSerTestReader::unsetSource() {
    d->context->disposeReader();
}

void CaSerTestReader::m_configure(const CuData& da) {
    QString description, unit, label;
    CuVariant m, M;
    m = da["min"];  // min value
    M = da["max"];  // max value
}


void CaSerTestReader::onUpdate(const CuData &da) {
    emit newData(d->creat_usecs, da);
}

