#ifndef CaSerTestReader_H
#define CaSerTestReader_H

#include <QObject>
#include <cudatalistener.h>
#include <cucontexti.h>
#include <cudata.h>

#include <QString>
class QContextMenuEvent;

class CaSerTestReaderPrivate;
class Cumbia;
class CumbiaPool;
class CuControlsReaderFactoryI;
class CuControlsFactoryPool;
class CuContext;
class CuLinkStats;

/*! \brief CaSerTestReader is a reader that uses QObject to display values.
 *
 * Connection is initiated with setSource. When new data arrives, it is displayed and the newData convenience
 * signal is emitted.
 * 
 * getContext returns a pointer to the CuContext used as a delegate for the connection.
 *
*/
class CaSerTestReader : public QObject, public CuDataListener, public CuContextI
{
    Q_OBJECT

public:
    CaSerTestReader(QObject *parent, CumbiaPool *cumbia_pool,
                    const CuControlsFactoryPool &fpool,
                    double creat_usecs, long id);

    virtual ~CaSerTestReader();

    /** \brief returns the source of the reader
     *
     * @return a QString with the name of the source
     */
    QString source() const;

    /** \brief returns a pointer to the CuContext used as a delegate for the connection.
     *
     * @return CuContext
     */
    CuContext *getContext() const;

    long creat_usecs() const;

    long id() const;

public slots:

    /** \brief set the source and start reading
     *
     * @param the name of the source
     */
    void setSource(const QString& s);

    /** \brief disconnect the source
     *
     * remove the source and stop reading
     */
    void unsetSource();

signals:
    void newData(double creat_us, const CuData&);

    void linkStatsRequest(QWidget *myself, CuContextI *myself_as_cwi);

protected:

private:
    CaSerTestReaderPrivate *d;

    void m_init();

    void m_configure(const CuData& d);


    // CuDataListener interface
public:
    void onUpdate(const CuData &d);
};

#endif // QUTLABEL_H
