#include "casertestreader.h"
#include "caserver-test-qt.h"
#include "ui_caserver-test-qt.h"

// cumbia
#include <cumbiapool.h>
#include <cuserviceprovider.h>
#include <cumacros.h>
#include <quapps.h>
#include <qustring.h>
#include <QSettings>
// cumbia

#include <QTimer>
#include <QTextStream>
#include <QFile>
#include <QCommandLineParser>
#include <QCommandLineOption>
#include <QRandomGenerator>
#include <QMessageBox>
#include <QtDebug>
#include <sys/time.h>

Caservertestqt_P::Caservertestqt_P() {
    subscribe_tmr = timer_sec = nullptr;
    counters[CounterSub] = counters[CounterUnsub] = 0;
    err_bg = QColor(Qt::red);
    err_bg.setAlpha(125);
}

void Caservertestqt_P::reset_counters() {
    counters[CounterSub] = counters[CounterUnsub] = 0;
}

Caservertestqt::Caservertestqt(CumbiaPool *cumbia_pool,
                               QWidget *parent,
                               const QString& srclistfile) :
    QWidget(parent)
{
    d = new Caservertestqt_P;
    d->srclistfile = srclistfile;

    // cumbia
    CuModuleLoader mloader(cumbia_pool, &m_ctrl_factory_pool, &m_log_impl);
    cu_pool = cumbia_pool;
    ui = new Ui::Caservertestqt;
    ui->setupUi(this, cu_pool, m_ctrl_factory_pool);
    m_load_settings(); // after setupUi

    // mloader.modules() to get the list of loaded modules
    // cumbia

    //
    ui->tabWidget->setCurrentIndex(0);
    foreach(QTreeWidget *w , findChildren<QTreeWidget *>())
        w->header()->setSectionResizeMode(QHeaderView::ResizeToContents);

    // load sources from file
    loadPool();

    connect(ui->pbReload, SIGNAL(clicked()), this, SLOT(loadPool()));
    connect(ui->pbStart, SIGNAL(clicked()), this, SLOT(start()));
    connect(ui->pbStop, SIGNAL(clicked()), this, SLOT(stop()));

    QList<QProgressBar *> pbs { ui->pbSub, ui->pbUnsub };
    foreach(QProgressBar *pb, pbs) {
        pb->setRange(0, ui->sbNew->value());
        pb->setVisible(false);
    }
}

Caservertestqt::~Caservertestqt()
{
    m_save_settings();
    delete d;
    delete ui;
}

void Caservertestqt::start() {
    if(d->subscribe_tmr)
        delete d->subscribe_tmr;
    if(ui->sbNew->value() > 0) {
        const int interval = 1000 / ui->sbNew->value();
        d->subscribe_tmr = new QTimer(this);
        d->subscribe_tmr->setInterval(interval);
        connect(d->subscribe_tmr, SIGNAL(timeout()), this, SLOT(onMainTimeout()));
        d->subscribe_tmr->start();
    }

    if(!d->timer_sec) {
        d->timer_sec = new QTimer(this);
        d->timer_sec->setInterval(1000);
        connect(d->timer_sec, SIGNAL(timeout()), this, SLOT(onTimeoutSec()));
    }
    d->timer_sec->start();
}

void Caservertestqt::stop() {
    if(d->subscribe_tmr)
        d->subscribe_tmr->stop();
    if(d->timer_sec)
        d->timer_sec->stop();
}

void Caservertestqt::onMainTimeout() {
    QRandomGenerator rndg;
    time_t tp;
    time(&tp);
    rndg.seed(tp);
    // pick the source from src pool tree
    int ra = rndg.bounded((qint32) 0, ui->twSrcPool->topLevelItemCount());
    QTreeWidgetItem *it = ui->twSrcPool->topLevelItem(ra);
    printf("Caservertestqt::onMainTimeout: creating reader %s\n", qstoc(it->text(1)));
    m_start_reader(it->text(1), it->text(2).toULong());
    m_stop_reader();
    d->counters[Caservertestqt_P::CounterSub]++;
    d->counters[Caservertestqt_P::CounterUnsub]++;

    // no need to update progress bar in cycle above
    ui->pbSub->setValue(d->counters[Caservertestqt_P::CounterSub]);
    ui->pbUnsub->setValue(d->counters[Caservertestqt_P::CounterUnsub]);
    if(d->counters[Caservertestqt_P::CounterSub] == (uint) ui->sbNew->value())
        d->counters[Caservertestqt_P::CounterSub] = 0;
    if(d->counters[Caservertestqt_P::CounterUnsub] == (uint) ui->sbNew->value())
        d->counters[Caservertestqt_P::CounterUnsub] = 0;

    QList<QProgressBar *> pbs { ui->pbSub, ui->pbUnsub };
    foreach(QProgressBar *pb, pbs) {
        pb->setRange(0, ui->sbNew->value());
        pb->setTextVisible(true);
        pb->setVisible(pb->value() != pb->minimum() && pb->value() != pb->maximum());
    }
}

void Caservertestqt::onTimeoutSec() {
    QTreeWidget *tw = ui->twRunning;
    const QDateTime now = QDateTime::currentDateTime();
    qint64 secs = now.currentSecsSinceEpoch();
    for(int i = 0; i < tw->topLevelItemCount(); i++) {
        // advance seconds counter in running view, "Running" column
        QTreeWidgetItem *it = tw->topLevelItem(i);
        it->setText(ColDataElapsed, QString::number(it->text(ColDataElapsed).toLong() +1));
        // advance seconds counter in last updated secs

        QDateTime last_u_dt = it->data(0, Data_Updated_DateTime).toDateTime();
        qint64 secs_since_last_update = last_u_dt.secsTo(now);
        it->setData(0, ColDataUpdated, QString("%1s ago").arg(secs_since_last_update));

        // hide old stopped entries
        const int rem_after = ui->sbRemoveStoppedAfter->value();
        if(it->isDisabled() &&
                secs - it->data(0, Data_Stop_Sec).toLongLong() > rem_after)
            delete it;
    }
}

// update view
void Caservertestqt::onNewData(double creat_us, const CuData &da) {
    QList<QTreeWidgetItem *>its = d->running_creat_src_map.values(creat_us);
    QDateTime now = QDateTime::currentDateTime();
    foreach(QTreeWidgetItem *it, its) {
        const QuString msg(da, "msg");
        const bool data_ok = !da.B("err");
        const QuString src(da, "src");
        const std::string& v = da["value"].toString();
        if(it->text(ColDataSrc) == src) {
            qint64 u_ms = da["timestamp_ms"].l();
            QDateTime tsms = QDateTime::fromMSecsSinceEpoch(u_ms);
            qint64 secs_since_last_update = tsms.secsTo(now);
            it->setData(0, ColDataUpdated, QString("%1s ago").arg(secs_since_last_update));
            bool ok = !it->data(0, DataOkFlag).isValid() ||
                    it->data(0, DataOkFlag).toBool();
            ok &= data_ok;
            it->setText(ColDataErr, !ok ? "YES" : "");
            if(!ok)
                it->setBackground(ColDataErr, d->err_bg);
            if(msg.length() > 0) // update with last msg if not empty
                it->setText(ColDataMsg, msg);
            it->setData(0, Data_Updated_DateTime, tsms);
            it->setText(ColDataUpdated, tsms.toString("hh:mm:ss"));
            if(ok) {
                const std::string& mode = da["mode"].s();
                const QString& old_mode = it->text(ColDataEvent);
                // mode changed ?
                if(old_mode != "-" && QString::fromStdString(mode) != old_mode)
                    it->setBackground(ColDataEvent, QColor("orange"));
                it->setText(ColDataValue, v.c_str());
                it->setText(ColDataEvent,  QuString(da, "mode"));
            }


            // child item with data
            QTreeWidgetItem *dit = new QTreeWidgetItem(it);
            QList<int> empty { ColDataSrc, ColDataStarted, ColDataElapsed, ColDataSrcId };
            foreach(const int &col, empty)
                dit->setText(col, "");
            dit->setText(ColDataUpdated, tsms.toString("hh:mm:ss"));
            dit->setText(ColDataMsg, msg);
            if(ok) {
                dit->setText(ColDataEvent, QuString(da, "mode")); // event (E) or poll (P)
                dit->setText(ColDataValue, v.c_str());
            }
        }
    }
}

CaSerTestReader *Caservertestqt::m_start_reader(const QString &src, ulong id) {
    struct timespec tv;
    clock_gettime(CLOCK_REALTIME, &tv);
    double ts = tv.tv_sec + tv.tv_nsec * 1e-6;
    CaSerTestReader *r = new CaSerTestReader(this, cu_pool, m_ctrl_factory_pool,ts, id);
    r->setObjectName(m_make_reader_nam(src, ts, id));
    printf("Caservertestqt::m_start_reader \e[1;32m reader %s with nam %s\n",
           qstoc(src), qstoc(r->objectName()));
    connect(r, SIGNAL(newData(double,CuData)), this, SLOT(onNewData(double,CuData)));
    r->setSource(src);
    QString dts = QDateTime::fromMSecsSinceEpoch(qRound64(ts * 1e3)).toString("hh:mm:ss.zzz");
    QTreeWidgetItem *it = new QTreeWidgetItem(ui->twRunning, QStringList() <<
                                              src<< dts<< "0" << "-" << "-" << "-" << "-" << "-" << QString::number(id));
    it->setData(0, Data_Creat_US, ts);
    it->setData(0, Data_Src_Id, static_cast<qlonglong>(id));
    it->setData(0, Data_Stop_Sec, -1);
    d->running_creat_src_map.insert(ts, it);

    return r;
}

void Caservertestqt::m_stop_reader() {
    if(ui->twRunning->topLevelItemCount() > ui->sbRunning->value()) {
        int i = 0;
        while(i < ui->twRunning->topLevelItemCount() && ui->twRunning->topLevelItem(i)->data(0, Data_Stop_Sec).toLongLong() >= 0)
            i++;
        if(i < ui->twRunning->topLevelItemCount()) {
            QTreeWidgetItem *it = ui->twRunning->topLevelItem(i);
            QString onam = m_make_reader_nam(it->text(0), it->data(0, Data_Creat_US).toDouble(),
                                             it->data(0, Data_Src_Id).toLongLong());
            CaSerTestReader *r = findChild<CaSerTestReader *>(onam);
            if(!r)
                perr("Caservertestqt.m_stop_reader: reader %s not found", qstoc(onam));
            r->unsetSource();
            it->setData(0, Data_Stop_Sec, QDateTime::currentSecsSinceEpoch());
            it->setDisabled(true);
            QTreeWidgetItem *finit = m_clone_item(it);
            ui->twFinished->insertTopLevelItem(ui->twFinished->topLevelItemCount(), finit);
            delete r;
        }
    }
}

QString Caservertestqt::m_make_reader_nam(const QString &src, double creat_us, long id) const {
    return QString("r_%1+%2+%3").arg(src).arg(creat_us, 0, 'f').arg(id);
}

// recursively clone item
QTreeWidgetItem *Caservertestqt::m_clone_item(const QTreeWidgetItem *it) const {
    qDebug() << __PRETTY_FUNCTION__;
    for(int i = 0;  it != nullptr && i < it->columnCount(); i++)
        printf(" %s / ", qstoc(it->text(i)));
    QTreeWidgetItem *nit = nullptr;
    if(it != nullptr) {
        nit = new QTreeWidgetItem(it->parent());
        for(int i = 0; i < it->columnCount(); i++) {
            nit->setText(i, it->text(i));
            nit->setToolTip(i, it->toolTip(i));
            for(int ur = Qt::UserRole + 1; ur < Qt::UserRole + 10; ur++)
                if(it->data(i, ur).isValid())
                    nit->setData(i, ur, it->data(i, ur));
        }
        for(int j = 0; j < it->childCount(); j++) {
            qDebug() << __PRETTY_FUNCTION__ << "cloning children " << it->childCount() << j;
            QTreeWidgetItem *chit = m_clone_item(it->child(j));
            if(chit) {
                nit->addChild(chit);
            }
            return chit;
        }
    }
    return nit;
}

void Caservertestqt::m_save_settings() {
    QSettings s(qApp->organizationName(), qApp->applicationName());
    foreach(QSpinBox *sb, findChildren<QSpinBox *>()) {
        s.setValue(sb->objectName(), sb->value());
    }
}

void Caservertestqt::m_load_settings() {
    QSettings s(qApp->organizationName(), qApp->applicationName());
    foreach(QSpinBox *sb, findChildren<QSpinBox *>()) {
        if(s.value(sb->objectName()).isValid()) {
            sb->setValue(s.value(sb->objectName()).toDouble());
        }
    }
}

void Caservertestqt::loadPool() {
    ui->twSrcPool->clear();
    d->srcmap.clear();
    QStringList filesrcs;
    QFile f(d->srclistfile);
    if(f.open(QIODevice::ReadOnly|QIODevice::Text)) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
        const qint32 &nsrcs_max = ui->sbSrcPoolCnt->value();
        QTextStream in(&f);
        int i = 0;
        while(!in.atEnd()) {
            i++;
            filesrcs << in.readLine();
        }
        QRandomGenerator rndg;
        time_t tp;
        time(&tp);
        rndg.seed(tp);
        int ra;
        d->filesrc_siz = filesrcs.size();
        for(int i = 0; i < nsrcs_max; i++) {
            ra = rndg.bounded((qint32) 0, d->filesrc_siz);
            d->srcmap.insert(ra, filesrcs[ra]);
        }

        i = 0;
        foreach(int id, d->srcmap.uniqueKeys()) {
            foreach(const QString &s, d->srcmap.values(id)) { // multi map
                QTreeWidgetItem *si =
                        new QTreeWidgetItem(ui->twSrcPool, QStringList() << QString::number(++i) << s << QString::number(id));
                si->setToolTip(0, QString("src \"%1\" file %2 line %3 ").arg(si->text(0)).arg(d->srclistfile).arg(id));
                if(d->srcmap.values(id).size() > 1)
                    si->setBackground(0, QColor(Qt::cyan));
            }
        }
        ui->lSrcs->setText(QString("Srcs: %1/%2").arg(ui->sbSrcPoolCnt->value()).arg(d->filesrc_siz));

        QApplication::restoreOverrideCursor();
    }
    else
        QMessageBox::critical(this, "Error opening file", QString("Error opening file %1 for reading: %2").arg(d->srclistfile).arg(f.errorString()));

}

